from sync.fields import AsymmetricFieldsSynchronizer

from youtrack.sync.comments import CommentSynchronizer


class AsymmetricIssueMerger(object):
    def __init__(self, master, slave, master_executor, slave_executor, issue_binder, link_synchronizer, fields_to_sync, last_run, current_run, project_id):
        self.master = master
        self.slave = slave
        self.master_executor = master_executor
        self.slave_executor = slave_executor
        self.last_run = last_run
        self.current_run = current_run
        self.comment_sync = CommentSynchronizer(master, slave, master_executor, slave_executor)
        self.field_sync = AsymmetricFieldsSynchronizer(master, slave, master_executor, slave_executor, fields_to_sync)
        self.issue_binder = issue_binder
        self.link_synchronizer = link_synchronizer
        self.project_id = project_id

    def sync(self, master_issue_id, slave_issue_id):
        _master_issue_id = master_issue_id
        _slave_issue_id =  slave_issue_id
        _result = None
        try:
            if not master_issue_id and slave_issue_id:
                _master_issue_id = self.issue_binder.slaveIssueIdToMasterIssueId(slave_issue_id)
                _result = _master_issue_id
            elif not slave_issue_id and master_issue_id:
                _slave_issue_id = self.issue_binder.masterIssueIdToSlaveIssueId(master_issue_id)
                _result = _slave_issue_id
            self._sync(_master_issue_id, _slave_issue_id, self.last_run, self.current_run)
        except KeyError as error:
            print(error)
        return _result

    def _sync(self, master_issue_id, slave_issue_id, last_run, current_run):
        self.field_sync.syncFields(master_issue_id, slave_issue_id, last_run, current_run)
        self.comment_sync.syncComments(master_issue_id, slave_issue_id)
        self.link_synchronizer.collectLinksToSyncById(master_issue_id, slave_issue_id)

    def clone_issue_to_master(self, issue_from):
        safe_summary = issue_from.summary if hasattr(issue_from, 'summary') else ''
        safe_description = issue_from.description if hasattr(issue_from, 'description') else ''
        created_issue_id = self.master_executor.createIssue(self.project_id, safe_summary, safe_description, issue_from.id)
        if created_issue_id: self._sync(created_issue_id, issue_from.id, None, self.current_run)
        return created_issue_id

    def clone_issue_to_slave(self, issue_from):
        safe_summary = issue_from.summary if hasattr(issue_from, 'summary') else ''
        safe_description = issue_from.description if hasattr(issue_from, 'description') else ''
        created_issue_id = self.slave_executor.createIssue(self.project_id, safe_summary, safe_description, issue_from.id)
        if created_issue_id: self._sync(issue_from.id, created_issue_id, None, self.current_run)
        return created_issue_id


from youtrack import YouTrackException

LOGGED_COMMENT_LENGTH = 10

class SafeCommandExecutor(object):
    def __init__(self, yt, logger):
        self.yt = yt
        self.logger = logger
        self.debug_mode = False

    def setDebugMode(self, on):
        self.debug_mode = on

    def executeCommand(self, issue_id, command, comment=None, run_as=None):
        if command != '':
            try:
                if not self.debug_mode:
                    self.yt.executeCommand(issue_id, command, comment=comment, run_as=run_as)
                if comment:
                    self.logger.logAction(issue_id, self.yt, 'added comment: \"' + comment[0:LOGGED_COMMENT_LENGTH] + '...\"', run_as)
                else:
                    self.logger.logAction(issue_id, self.yt, 'applied command: \"' + command + '\"', run_as)
            except Exception as e:
                self.logger.logError(e, issue_id, self.yt, 'failed to apply command: \"' + command + '\"', run_as)

    def executeUserImport(self, user):
        if user:
            try:
                if not self.debug_mode:
                    self.yt.importUsers([user])
                self.logger.logAction('Import user', self.yt, 'imported user: \"' + str(user.login) + '\"')
            except YouTrackException as e:
                self.logger.logError(e, 'Import user', self.yt, 'failed to import user: \"' + user.login + '\" - could not find in opposite youtrack')

    def createIssue(self, project_id, summary, description, issue_from_id):
        try:
            created_issue_id = 'DEBUG'
            if not self.debug_mode:
                created_message = self.yt.createIssue(project_id, None, summary, description)
                created_issue_id = project_id + '-' + created_message.rpartition('-')[2]
                #fail if summary or description are too long or can't convert to unicode params
            self.logger.logAction(str(issue_from_id) + '->' + created_issue_id, self.yt, 'created')
            return None if self.debug_mode else created_issue_id
        except Exception as e:
            self.logger.logError(e, str(issue_from_id) + '->?', self.yt, 'failed to create' )
            return None

    def importLinks(self, links, permitted_issue_ids):
        maxLinks = 100
        links_to_import = []
        for link in links:
            if link.target not in permitted_issue_ids:
                message = 'failed to import link ' + self._getPrettyLink(link)  + ' because ' + link.target + ' was not imported'
                self.logger.logError(None, 'Links', self.yt, message)
            elif link.source not in permitted_issue_ids:
                message = 'failed to import link ' + self._getPrettyLink(link) + ' because ' + link.source + ' was not imported'
                self.logger.logError(None, 'Links', self.yt, message)
            else:
                links_to_import.append(link)
                if len(links_to_import) == maxLinks:
                    self._import_links_batch(links_to_import)
                    links_to_import = []
        if len(links_to_import):
            self._import_links_batch(links_to_import)

    def _import_links_batch(self, links_to_import):
        if not self.debug_mode:
            self.yt.importLinks(links_to_import)
        for link in links_to_import:
            message = 'imported ' + self._getPrettyLink(link)
            self.logger.logAction('Links', self.yt, message)

    def _getPrettyLink(self, link):
           return link.typeName + ' link: ' + link.source + '->' + link.target

    def getLogger(self):
        return self.logger


import youtrack

fix = "fix"
opn = "open"
rop = "reopen"
inp = "in progress"
dis = "discuss"
crp = "can't reproduce"
obs = "obsolete"
dup = "duplicate"
asd = "as designed"
inv = "invalidate"
wat = "wait"
ver = "verify"
cvr = "can't verify"

advanced_state_machine = {
    "Submitted -> Fixed" : fix,
    "Submitted -> Open" : opn,
    "Submitted -> In Progress" : inp,
    "Submitted -> To be discussed" : dis,
    "Submitted -> Can't Reproduce" : crp,
    "Submitted -> Obsolete" :obs,
    "Submitted -> Duplicate" :dup,
    "Submitted -> As designed" : asd,
    "Submitted -> Invalid" : inv,
    "Open -> In Progress" : inp,
    "Open -> To be discussed" : dis,
    "Open -> Fixed" : fix,
    "Open -> Obsolete" : obs,
    "Open -> Duplicate" : dup,
    "Open -> Can't Reproduce" : crp,
    "Open -> As designed" : asd,
    "Open -> Invalid" : inv,
    "Open -> Wait for Reply" : wat,
    "Reopened -> Open" : opn,
    "Obsolete -> Open" : rop,
    "Duplicate -> Open" : rop,
    "In Progress -> Open" : rop,
    "In Progress -> Fixed" : fix,
    "In Progress -> Can't Reproduce" : crp,
    "In Progress -> Obsolete" : obs,
    "In Progress -> As designed" : asd,
    "To be discussed -> In Progress" : inp,
    "To be discussed -> Duplicate" : dup,
    "To be discussed -> Obsolete" : obs,
    "Can't Reproduce -> Open" : rop,
    "As designed -> Open" : rop,
    "Won't fix -> Open" : rop,
    "Invalid -> Open" : rop,
    "Incomplete -> Open" : rop,
    "Fixed -> Open" : rop,
    "Fixed -> Verified" : ver,
    "Fixed -> W/O verification" : cvr,
    "W/O verification -> Open" : ver,
    "Verified -> Open" : rop,
    "Wait for Reply -> Fixed" : fix,
    "Wait for Reply -> Open" : opn,
    "Wait for Reply -> In Progress" : inp,
    "Wait for Reply -> To be discussed" : dis,
    "Wait for Reply -> Obsolete" : obs,
    "Wait for Reply -> Duplicate" : dup,
    "Wait for Reply -> As designed" :asd,
    "Wait for Reply -> Invalid" : inv
}

def get_event(field):
    old = field.old_value[0] if len(field.old_value) == 1 else None
    new = field.new_value[0] if len(field.new_value) == 1 else None
    if not old or not new : raise ValueError('State can not have multiple value')
    event = advanced_state_machine.get(old + ' -> ' + new)
    if not event: raise LookupError("failed to apply change: State:" + old + "->" + new + " - state machine doesn't allow this transition")
    return event

def get_command_for_state_change(field, with_state_machine):
    return "State " + (get_event(field) if with_state_machine else field.new_value[0]) + " "


from os import path
from setuptools import setup

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.md'), 'r') as f:
    long_description = f.read()

# Get version from file
with open(path.join(here, 'version'), 'r') as f:
    version = f.read().strip()

setup(
    name='youtrack',
    version=version,
    python_requires='>3',
    packages=['youtrack', 'youtrack.sync'],
    url='https://github.com/JetBrains/youtrack-rest-python-library',
    license='Apache 2.0',
    maintainer='Alexander Buturlinov',
    maintainer_email='imboot85@gmail.com',
    description='Python library for interacting with YouTrack via REST API supporting python3',
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        'httplib2 >= 0.7.4',
        'six'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent"
    ],
)



import sys
import time

from youtrack import YouTrackException
from youtrack.sync.states import get_event

PRIORITY_MAPPING= {'0':'Show-stopper', '1':'Critical', '2':'Major', '3':'Normal', '4':'Minor'}

def get_issue_changes(yt, issue_id, start=None, finish=None):
     result = yt.get_changes_for_issue(issue_id)
     after_ms = get_in_milliseconds(start) if start else 0
     before_ms = get_in_milliseconds(finish) if finish else sys.maxsize
     new_changes = []
     for change in result:
         change_time = change['updated']
         if (change_time > after_ms) and (change_time < before_ms):
             new_changes.append(change)
     return new_changes

def get_in_milliseconds(_datetime):
    return int(round(1e+3*time.mktime(_datetime.timetuple()) + 1e-3*_datetime.microsecond))

class AsymmetricFieldsSynchronizer(object):
    def __init__(self, master, slave, master_executor, slave_executor, fields_to_sync):
        self.master = master
        self.slave = slave
        self.executors = {master : master_executor, slave : slave_executor}
        self.fields_to_sync = fields_to_sync

    def syncFields(self, master_issue_id, slave_issue_id, last_run, current_run):
        #sync fields
        slave_changes = get_issue_changes(self.slave, slave_issue_id, last_run, current_run)
        master_changes = get_issue_changes(self.master, master_issue_id, last_run, current_run)
        #field changes made in master should rewrite any field changes in slave
        changed_fields = self._apply_changes_to_issue(self.slave, self.master, slave_issue_id, master_changes)
        self._apply_changes_to_issue(self.master, self.slave, master_issue_id, slave_changes, fields_to_ignore=changed_fields)

    def _apply_changes_to_issue(self, to_yt, from_yt, issue_id, changes, fields_to_ignore=None):
        if not fields_to_ignore: fields_to_ignore = []
        changed_fields = set()
        for change in changes:
            command, run_as = self._convert_change_to_command(issue_id, change, changed_fields, fields_to_ignore, from_yt, to_yt)
            self.executors[to_yt].executeCommand(issue_id, command, run_as=run_as)
        return changed_fields

    def _convert_change_to_command(self, issue_id, change, changed_fields, fields_to_ignore, from_yt, to_yt):
        run_as = change.updater_name
        self._try_to_sync_user(to_yt, from_yt, run_as)
        command = ''
        for field in change.fields:
            field_name = field.name.lower()
            if field.name != 'links' and field_name in self.fields_to_sync and field_name not in fields_to_ignore:
                try:
                    command += self._create_command(field, field_name, to_yt == self.master)
                    changed_fields.add(field_name)
                except Exception as error:
                    self.executors[to_yt].getLogger().logError(error, issue_id, to_yt, error.message)
        return command, run_as

    def _create_command(self, field, field_name, in_master):
        new_value = [get_event(field)] if in_master and field_name == 'state' else field.new_value
        return self.get_command_set_value_to_field(field_name, new_value)

    def get_command_set_value_to_field(self, field_name, new_value):
        command = ""
        for new_field_value in new_value:
            if new_field_value:
                if field_name == 'priority' and new_field_value in list(PRIORITY_MAPPING.keys()):
                    new_field_value = PRIORITY_MAPPING[new_field_value]
                command += field_name + " " + new_field_value + " "
        return command

    def _try_to_sync_user(self, to_yt, from_yt, login):
        try:
            to_yt.getUser(login)
        except YouTrackException:
            user_to_import = from_yt.getUser(login)
            self.executors[to_yt].executeUserImport(user_to_import)
